structure LambdaAst : LAMBDA_AST = struct

  type var = string * int

  fun eqvar (s1, i1) (s2, i2) = i1 = i2

  datatype primop = 
    Plus                   
  | Times
  | Minus
  | ZeroP
  | Fst
  | Snd

  datatype expr =
    Var of var
  | Num of int
  | Op of primop
  | Lambda of var list * expr
  | App of expr list
  | Pair of expr * expr
  | Case of expr * (var * expr) * (var * expr)
  | Left of expr
  | Right of expr
  | Letrec of (var * expr) list * expr


  exception Bad
  val vctr = ref 0
  fun newvar s = let 
    val v = !vctr 
    val _ = vctr := v+1
  in
    (s, v)
  end

  fun alpha_vary expr =
    let
      fun add ctxt str i = (str, i)::ctxt
      fun lookup x [] = raise Bad
	| lookup x ((str, i)::rest) = if x = str then (str, i) else
	lookup x rest
      fun bindnew vl ctxt=
	let val (vl', ctxt') = 
	  List.foldr 
	  (fn ((str,_), (vl', ctxt)) => 
	   let val (_, i) = newvar str in
	     ((str, i)::vl', add ctxt str i)
	   end)
	  ([], ctxt) vl
	in
	  (vl', ctxt')
	end
      
      fun alpha ctxt e =
	case e of
	  Var (str, i) => (Var(lookup str ctxt) handle Bad => Var(str, i))
	| Lambda (vl, e) => 
	    let val (vl', ctxt') = bindnew vl ctxt in
	      Lambda(vl', alpha ctxt' e)
	    end
	| App el => App(List.map (alpha ctxt) el)
	| Pair(e1,e2) => Pair((alpha ctxt) e1, (alpha ctxt) e2)
	| Letrec (vel, e) => 
	    let 
	      val (vl, el) = ListPair.unzip vel
	      val (vl', ctxt') = bindnew vl ctxt
	      val el' = List.map (alpha ctxt') el
	      val (vel') = ListPair.zip (vl', el')
	    in
	      Letrec(vel', alpha ctxt' e)
	    end
	| Left e => Left ((alpha ctxt) e)
	| Right e => Right ((alpha ctxt) e)
	| Case (e,(v1,e1),(v2,e2)) => 
	    let
	      val e' = alpha ctxt e
	      val ([v1'],ctxt1) = bindnew [v1] ctxt
	      val e1' = alpha ctxt1 e1
	      val ([v2'],ctxt2) = bindnew [v2] ctxt
	      val e2' = alpha ctxt2 e2
	    in Case(e',(v1',e1'),(v2',e2'))
	    end
	| e => e
    in
      alpha [] expr
    end


  fun alpha_equiv expr =
    let
      fun add ctxt i j = (i, j)::ctxt
      fun lookup x [] = raise Bad
	| lookup x ((i, j)::rest) = if x = i then j else
	lookup x rest
      fun bindnew vl1 vl2 ctxt =
	List.foldr 
	  (fn (((_,i),(_,j)), (ctxt)) => 
	    add ctxt i j)
	  ctxt (ListPair.zip (vl1, vl2))
      
      fun alpha ctxt e1 e2 =
	case (e1,e2) of
	  (Var (_, i),Var(_,j)) => 
	    ((lookup i ctxt = j) handle Bad => false)
	| (Lambda (vl1, e1),Lambda(vl2,e2)) => 
	    let val ctxt' = bindnew vl1 vl2 ctxt in
	      alpha ctxt' e1 e2 
	    end
	| (App el1,App el2) => 
	    List.foldl (fn ((x1,x2),y) => (alpha ctxt x1 x2) andalso y)
	      true
	      (ListPair.zip (el1,el2))
	| (Pair(e1,e1'),Pair(e2,e2')) => 
	      alpha ctxt e1 e2 andalso alpha ctxt e1' e2' 
	| (Letrec (vel1, e1),Letrec(vel2,e2)) => 
	    let 
	      val (vl1, el1) = ListPair.unzip vel1
	      val (vl2,el2) = ListPair.unzip vel2
	      val (ctxt') = bindnew vl1 vl2 ctxt
	    in
	      List.foldl 
	          (fn ((x1,x2),y) => (alpha ctxt x1 x2) andalso y)
		  true
		  (ListPair.zip (el1,el2))
	    end
	| (Left e1,Left e2) => alpha ctxt e1 e2
	| (Right e1, Right e2) => alpha ctxt e1 e2
	| (Case (e1,(v11,e11),(v12,e12)),Case(e2,(v21,e21),(v22,e22))) => 
	    alpha ctxt e1 e2 
	    andalso
	    alpha (bindnew [v11] [v21] ctxt) e11 e21
	    andalso
	    alpha (bindnew [v12] [v22] ctxt) e12 e22
	| (e1,e2) => e1=e2
    in
      alpha [] expr
    end


  val astTrue = Left(Num 0)

  val astFalse = Right(Num 0)
    
  val astIf = 
    let 
      val e = newvar "e";
      val e1 = newvar "e1";
      val e2 = newvar "e2";
    in Lambda([e,e1,e2],Case(Var e,(newvar"t",Var e1),(newvar"f",Var e2)))
    end

  val astError = App[Num 0, Num 0]

  val astHead = 
    let val l = newvar "l";
      val h = newvar "h";
      val err = newvar "error";
    in
      Lambda([l],Case(Var l,(err,astError),
		      (h,App [(Op Fst),Var h])))
    end

  val astTail = 
    let val l = newvar "l";
      val t = newvar "t";
      val err = newvar "error";
    in
      Lambda([l],Case(Var l,(err,astError),
		      (t,App[Op Snd,Var t])))
    end

  val astNilP = 
    let val l = newvar "l"
    in 
      Lambda([l],App [astIf, Var l, astTrue, astFalse])
    end

  val astCons = 
    let
      val h = newvar "h";
      val t = newvar "t";
    in 
      Lambda([h,t],Right(Pair(Var h, Var t)))
    end

  val astNil = Left(Num 0)
end
