structure LambdaInterpretation : LAMBDA_INTERPRETATION = struct
    structure LambdaAst = LambdaAst
    structure PP = LambdaPPrint

    exception RuntimeError of string

    local open LambdaAst in
    fun subst e v e' =
      case e' of
	Var v' => if eqvar v v' then e else e'
      | Num i => Num i
      | Op p => Op p
      | Lambda (vl, expr) => (* Assumes alpha-conversion, 
			      so capture is not possible *)
	  Lambda (vl, subst e v expr)
      | App el => App (List.map (subst e v) el)
      | Pair(e1,e2) => Pair(subst e v e1, subst e v e2)
      | Case(e',(v1,e1),(v2,e2)) => Case(subst e v e', 
					 (v1,subst e v e1), 
					 (v2,subst e v e2))
      | Left(e') => Left(subst e v e')
      | Right(e') => Right(subst e v e')
      | Letrec (vel, expr) => 
	  let 
	    val (vs, es) = ListPair.unzip vel 
	    val es' = List.map (subst e v) es
	    val vel' = ListPair.zip (vs, es')
	  in
	    Letrec (vel', subst e v expr)
	  end

    fun interpret e = 
      let
	fun interpPrim p args =
	  let 
	    fun evalToInt l =
	      List.map (fn e => case interp e of
			Num i => i
		      | _ => (PP.ppE e; raise (RuntimeError ("+ or * applied to non-int"))))
	      l
	  in
	    case p of
	      Plus => Num(List.foldl (op +) 0 (evalToInt args))
	    | Times => Num(List.foldl (op * ) 1 (evalToInt args))
	    | Minus => (case args of
			  [e1, e2] => 
			    let 
			      val v1 = interp e1
			      val v2 = interp e2
			    in
			      case (v1, v2) of
				(Num n1, Num n2) => Num(n1-n2)
			      | _ => raise (RuntimeError 
					    ("- applied to bad args"))
			    end
			| _ => raise 
			    (RuntimeError ("- applied to wrong number of args")))
	    | ZeroP => (case args of 
			  [e] => let val v = interp e in
			    case v of
			      Num 0 => astTrue
			    | _ => astFalse
			  end
			| _ => raise (RuntimeError ("zero? applied to wrong number of args")))
	    | Fst => (case args of 
			[e] => let val v = interp e in 
			  case v of
			    Pair(e1,e2) => let val v' = interp e1 in v' end
			  | _ => raise (RuntimeError ("fst applied to non-pair"))
			       end
		      | _ => raise (RuntimeError ("fst applied to wrong number of arguments")))
	    | Snd => (case args of 
			[e] => let val v = interp e in 
			  case v of
			    Pair(e1,e2) => let val v' = interp e2 in v' end
			  | _ => raise (RuntimeError ("snd applied to non-pair"))
			       end
		      | _ => raise (RuntimeError ("snd applied to wrong number of arguments")))
	  end

	and interp e =
	  case e of
	    Var v => Var v
	  | App [] => raise (RuntimeError "Null Application")
	  | App (f::args) => 
	      let val f = interp f in
		case f of
		  Op p => interpPrim p args
		| Lambda (vars, e) => 
		    (case (vars, args) of 
		      ([], []) =>
			interp e
		    | (vs, []) => 
			Lambda (vs, e)
		    | (v1::vrest, a1::arest) =>
			let
			  val e' = subst a1 v1 (alpha_vary e) 
			in
			  interp (App ((Lambda (vrest, e')):: arest)) 
			end
		    | _ => raise (RuntimeError 
				"fn applied to wrong number of args"))
		| e => (PP.ppE e;
		       raise (RuntimeError "Applied non-function to args"))
	      end
	  | Letrec (vel, e) => 
	      let
		val e' = List.foldl (fn ((v, a), e) =>
				 subst a v e)
		  e vel
		val e'' = List.foldl (fn ((v, _), e) =>
				      subst (alpha_vary (Letrec (vel, Var v))) v e)
		  e' vel
	      in
		interp e''
	      end
	  | Case(e,(v1,e1),(v2,e2)) => 
	      let val s = interp e 
	      in case s of
		Left(e') => interp (subst (alpha_vary e') v1 e1)
	      | Right (e') => interp (subst (alpha_vary e') v2 e2)
		| _ => raise (RuntimeError "Case applied to non [left/right]")
	      end
	  | e => e
      in
	(interp e) handle (RuntimeError s) => (print (s^"\n"); App [])
      end


    fun interpret2 e = 
      let       
	fun interpPrim p args =
	  let 
	    fun evalToInt l =
	      List.map (fn e => case interp e of
			Num i => i
		      | _ => (PP.ppE e; raise (RuntimeError ("+ or * applied to non-int"))))
	      l
	  in
	    case p of
	      Plus => Num(List.foldl (op +) 0 (evalToInt args))
	    | Times => Num(List.foldl (op * ) 1 (evalToInt args))
	    | Minus => (case args of
			  [e1, e2] => 
			    let 
			      val v1 = interp e1
			      val v2 = interp e2
			    in
			      case (v1, v2) of
				(Num n1, Num n2) => Num(n1-n2)
			      | _ => raise (RuntimeError 
					    ("- applied to bad args"))
			    end
			| _ => raise 
			    (RuntimeError ("- applied to wrong number of args")))
	    | ZeroP => (case args of 
			  [e] => let val v = interp e in
			    case v of
			      Num 0 => LambdaTranslate.transTrue
			    | _ => LambdaTranslate.transFalse
			  end
			| _ => raise (RuntimeError ("zero? applied to wrong number of args")))

	    | (Fst|Snd) => raise (RuntimeError "Unsupported primop")
	  end

	and interp e =
	  case e of
	    Var v => Var v
	  | App [] => raise (RuntimeError "Null Application")
	  | App (f::args) => 
	      let val f = interp f in
		case f of
		  Op ZeroP => interpPrim ZeroP args
		| Op p => (if List.length args <> 2 then
			     raise (RuntimeError "primop not applied to 2 args")
			   else
			     interpPrim p args)
		| Lambda (vars, e) => 
		    if (List.length vars = 1) andalso 
		       (List.length args = 1) then
		      let 
			val e' = 
			  List.foldl (fn ((v, a), e) =>
				      subst a v (alpha_vary e)) 
			  e
			  (ListPair.zip (vars, args))
		      in
			interp e'
		      end
		    else raise (RuntimeError 
				"fn applied to wrong number of args")
		| e =>(PP.ppE e;
		       raise (RuntimeError "Applied non-function to args"))
	      end
	  | Letrec (vel, e) => 
	      raise (RuntimeError "Letrec not supported")
	  | Num n => Num n
	  | Op p => (case p of 
		       (Plus | Times | Minus | ZeroP) => Op p
		     | _ => raise (RuntimeError "Op not supported"))
	  | Lambda (vl, e) => 
		if List.length vl <> 1 then
		  (raise (RuntimeError "Lambda has more than one arg"))
		else Lambda (vl, e)
	  | Case (e,(v1,e1),(v2,e2)) => raise (RuntimeError "Case not supported")
	  | Pair (e1,e2) => raise (RuntimeError "Pair not supported")
	  | Left e => raise (RuntimeError "Left not supported")
	  | Right e => raise (RuntimeError "Right not supported")
      
      in
	(interp e) handle (RuntimeError s) => (print (s^"\n"); App [])
      end
   end
end
