structure LambdaSimpleAst : LAMBDA_SIMPLE_AST = struct
  type var = LambdaAst.var

  val eqvar = LambdaAst.eqvar
  val newvar = LambdaAst.newvar

  datatype primop = Plus | Times | Minus
  datatype expr = Var of var
                | Num of int
                | ZeroP of expr
                | IsZero of expr
                | Lambda of var * expr
                | App of expr * expr
                | Op of primop * expr * expr
                | True
                | False
                | If of expr * expr * expr

  exception Bad
  fun alpha_vary expr =
    let
      fun add ctxt str i = (str, i)::ctxt
      fun lookup x [] = raise Bad
	| lookup x ((str, i)::rest) = if x = str 
					then (str, i) 
					else lookup x rest
      fun bindnew (str,_) ctxt = let
				   val (_, i) = newvar str
				 in
				   ((str, i), add ctxt str i)
				 end
      
      val rec alpha 
	= fn ctxt
	   => (fn Var (str, i) 
	        => (Var(lookup str ctxt) handle Bad => Var(str, i))
		| Num i => Num i
	        | ZeroP(e) => ZeroP(alpha ctxt e)
	        | IsZero(e) => IsZero(alpha ctxt e)
	        | Lambda (v, e)
	        => let
		     val (v', ctxt') = bindnew v ctxt
		   in
		     Lambda(v', alpha ctxt' e)
		   end
		| App (e1, e2)
		=> App (alpha ctxt e1, alpha ctxt e2)
		| Op (p, e1, e2)
		=> Op (p, alpha ctxt e1, alpha ctxt e2)
                | True => True
                | False => False
                | If (b,e1,e2) 
                => If (alpha ctxt b, alpha ctxt e1, alpha ctxt e2))
    in
      alpha [] expr
    end
  fun alpha_equiv expr =
    let
      fun add ctxt i j = (i, j)::ctxt
      fun lookup x [] = raise Bad
	| lookup x ((i, j)::rest) = if x = i 
				      then j 
				      else lookup x rest
      fun bindnew (_,i) (_,j) ctxt = add ctxt i j

      fun alpha ctxt e1 e2
	= (case (e1, e2)
	     of (Var (_, i), Var (_,j))
              => ((lookup i ctxt = j) handle Bad => false)
	      | (Num i, Num j) => i = j
	      | (ZeroP(e1), ZeroP(e2)) => alpha ctxt e1 e2
	      | (IsZero(e1), IsZero(e2)) => alpha ctxt e1 e2
	      | (Lambda (v1,e1), Lambda (v2,e2))
	      => let
		   val ctxt' = bindnew v1 v2 ctxt
		 in
		   alpha ctxt' e1 e2
		 end
	      | (App (e11,e12), App(e21,e22))
	      => (alpha ctxt e11 e21) andalso
                 (alpha ctxt e12 e22)
	      | (Op (p1,e11,e12), Op (p2,e21,e22))
	      => (p1 = p2) andalso
                 (alpha ctxt e11 e21) andalso
                 (alpha ctxt e12 e22)
	      | (True, True) => true
	      | (False, False) => true
	      | (If (b1,e11,e12), If (b2,e21,e22))
	      => (alpha ctxt b1 b2) andalso
                 (alpha ctxt e11 e21) andalso
                 (alpha ctxt e12 e22) 
              | _ => false)
    in
      alpha [] expr
    end

  exception convertFailure of string
  val rec convert
    = fn LambdaAst.Var v => Var v
       | LambdaAst.Num n => Num n
       | LambdaAst.Lambda ([v],e) => Lambda(v, convert e)
       | LambdaAst.Lambda _ => raise convertFailure 
                                     "Lambda has > 1 arguments"
       | LambdaAst.App [e1,e2] => App(convert e1, convert e2)
       | LambdaAst.App [e0,e1,e2]
       => let
	    val primop = case e0
			   of LambdaAst.Op p
			    => (case p
				  of LambdaAst.Plus => Plus
				   | LambdaAst.Times => Times
				   | LambdaAst.Minus => Minus
				   | _ => raise convertFailure
				                "App head is not arithmetic")
                            | _ => raise convertFailure
                                         "App head is not primop"
	  in
	    Op (primop, convert e1, convert e2)
	  end
       | LambdaAst.App _ => raise convertFailure
                                  "App has > 2 expressions"
       | LambdaAst.Op LambdaAst.ZeroP 
       => let
	    val z = newvar "z"
	  in 
	    Lambda(z, ZeroP(Var z))
	  end
       | LambdaAst.Op _ => raise convertFailure
                                 "Op not in simple language"
       | LambdaAst.Pair _ => raise convertFailure
                                   "Pair not in simple language"
       | LambdaAst.Case _ => raise convertFailure
                                   "Case not in simple language"
       | LambdaAst.Left _ => raise convertFailure
                                   "Left not in simple language"
       | LambdaAst.Right _ => raise convertFailure
                                    "Right not in simple language"
       | LambdaAst.Letrec _ => raise convertFailure
				     "Letrec not in simple language"
  val rec thunk
    = fn Num n => Num n
       | ZeroP e => ZeroP (thunk e)
       | IsZero e => IsZero (thunk e)
       | Op (primop, e1, e2) => Op (primop, thunk e1, thunk e2)
       | True => True
       | False => False
       | If (e0, e1, e2) => If (thunk e0, thunk e1, thunk e2)
       | Var x => App(Var x, Num 0)
       | Lambda(x, e) => Lambda(x, thunk e)
       | App(e1, e2) => let
			  val u = newvar "u"
			in 
			  App(thunk e1, Lambda(u, thunk e2))
			end
  val fromLambdaAst = thunk o convert
  val transTrue = fromLambdaAst (LambdaTranslate.transTrue)
  val transFalse = fromLambdaAst (LambdaTranslate.transFalse)

  val rec toLambdaAst
    = fn Var v => LambdaAst.Var v
       | Num i => LambdaAst.Num i
       | ZeroP(e) 
       => LambdaAst.App [LambdaAst.Op LambdaAst.ZeroP,
			 toLambdaAst e]
       | IsZero(e) 
       => LambdaAst.App [toLambdaAst 
			 (fromLambdaAst 
			  (LambdaAst.Op LambdaAst.ZeroP)), 
			 toLambdaAst e]
       | Lambda (v,e) => LambdaAst.Lambda ([v], toLambdaAst e)
       | App (e1,e2) => LambdaAst.App [toLambdaAst e1, toLambdaAst e2]
       | Op (p,e1,e2) => let
			   val primop = case p
					  of Plus => LambdaAst.Plus
					   | Times => LambdaAst.Times
					   | Minus => LambdaAst.Minus
			 in
			   LambdaAst.App [LambdaAst.Op primop,
					  toLambdaAst e1,
					  toLambdaAst e2]
			 end
       | True => toLambdaAst (fromLambdaAst LambdaTranslate.transTrue)
       | False => toLambdaAst (fromLambdaAst LambdaTranslate.transFalse)
       | If (b,e1,e2)
       => LambdaAst.App
	  [LambdaAst.App
	   [LambdaAst.App 
	    [LambdaTranslate.translate LambdaAst.astIf,
	     toLambdaAst b],
	    toLambdaAst e1],
	   toLambdaAst e2]

  val ftLambdaAst = toLambdaAst o fromLambdaAst
end
